/*
 *	Ohio Trollius
 *	Copyright 1997 The Ohio State University
 *	RBD/GDB
 *
 *	$Id: waitany.c,v 6.2 1998/05/26 05:02:15 lam Exp $
 *
 *	Function:	- wait for first request to finish
 *	Accepts:	- # of requests
 *			- array of requests
 *			- index of completed request (out)
 *			- status (out)
 *	Returns:	- MPI_SUCCESS or error code
 *			- in the case of an error return index is set to
 *			  MPI_UNDEFINED if the error is not associated with
 *			  the completed request, e.g. an internal error
 */

#include <blktype.h>
#include <mpi.h>
#include <mpisys.h>
#include <mpitrace.h>
#include <rpisys.h>


int
MPI_Waitany(count, reqs, index, stat)

int			count;
MPI_Request		*reqs;
int			*index;
MPI_Status		*stat;

{
	int		i;
	int		nactive;		/* # active requests */
	int		errstat;		/* error in status */
	int		err;			/* error code */
	MPI_Comm	comm;			/* communicator */
	MPI_Request	req;			/* request */
	int		fl_trace;		/* do tracing? */
	double		startt;			/* start time */
	double		finisht;		/* finish time */

	lam_initerr_m();
	lam_setfunc_m(BLKMPIWAITANY);

	if (count < 0) {
		return(lam_errfunc(MPI_COMM_WORLD, BLKMPIWAITANY, 
				lam_mkerr(MPI_ERR_COUNT, 0)));
	}

	if ((count > 0) && (reqs == 0)) {
		return(lam_errfunc(MPI_COMM_WORLD, BLKMPIWAITANY,
				lam_mkerr(MPI_ERR_ARG, 0)));
	}

	if ((index == 0)) {
		return(lam_errfunc(MPI_COMM_WORLD, BLKMPIWAITANY,
				lam_mkerr(MPI_ERR_ARG, 0)));
	}

	if ((fl_trace = LAM_TRACE_TOP())) {
		startt = MPI_Wtime();
		_kio.ki_blktime = 0.0;
	}

	if (stat != MPI_STATUS_IGNORE) {
		lam_emptystat(stat);
	}
/*
 * While no request is done.
 */
	while (1) {
/*
 * Loop over the requests.
 */
		_mpi_req_blkclr_m();

		for (i = 0, nactive = 0; i < count; ++i) {
			req = reqs[i];
/*
 * Skip inactive requests.
 */
			if ((req == MPI_REQUEST_NULL) ||
					(req->rq_state == LAM_RQSINIT))
					continue;

			if (req->rq_state == LAM_RQSDONE) break;
/*
 * It's an active request.
 */
			++nactive;
			_mpi_req_blkset_m(req);
		}
/*
 * We found a completed request.
 */
		if (i < count) {
			*index = i;
			break;
		}
/*
 * If no active requests try to progress and we're done.
 */
		else if (nactive == 0) {
			*index = MPI_UNDEFINED;

			err = _mpi_req_advance();
			if (err != MPI_SUCCESS) {
				*index = MPI_UNDEFINED;
				return(lam_errfunc(MPI_COMM_WORLD,
						BLKMPIWAITANY, err));
			}
			
			lam_resetfunc_m(BLKMPIWAITANY);
			return(MPI_SUCCESS);
		}
/*
 * Otherwise block on any and loop again.
 */
		else {
			err = _mpi_req_advance();
			if (err != MPI_SUCCESS) {
				*index = MPI_UNDEFINED;
				return(lam_errfunc(MPI_COMM_WORLD,
						BLKMPIWAITANY, err));
			}
		}
	}
/*
 * Finish the completed request.
 */
	_mpi_req_rem_m(req);

	err = _mpi_req_end(req);
	if (err != MPI_SUCCESS) {
		return(lam_errfunc(MPI_COMM_WORLD, BLKMPIWAITANY, err));
	}

	if (stat != MPI_STATUS_IGNORE) {
		*stat = req->rq_status;
	}
	comm = req->rq_comm;
	errstat = req->rq_status.MPI_ERROR;
/*
 * Generate a run time trace except in the case of cancelled request or
 * a request where the peer is MPI_PROC_NULL.
 */
	if (fl_trace && (req->rq_rank != MPI_PROC_NULL)
			&& !(req->rq_flags & LAM_RQFCANCEL)) {
		finisht = MPI_Wtime();

		lam_tr_msg(
			(req->rq_type == LAM_RQIRECV) ? TRTINPUT : TRTNOIO,
			startt, LAM_S2US(finisht - startt - _kio.ki_blktime),
			LAM_S2US(_kio.ki_blktime), req->rq_rank, req->rq_tag,
			req->rq_comm, req->rq_dtype, req->rq_count,
			req->rq_status.MPI_SOURCE, req->rq_status.MPI_TAG,
			req->rq_seq, req->rq_type);
	}
/*
 * Reset persistent requests.
 * Destroy requests that are not persistent.
 */
	if (req->rq_marks & LAM_RQFPERSIST) {
		req->rq_state = LAM_RQSINIT;
	} else {
		err = _mpi_req_destroy(&reqs[i]);
		if (err != MPI_SUCCESS) {
			return(lam_errfunc(MPI_COMM_WORLD,
					BLKMPIWAITANY, err));
		}
	}

	if (errstat != MPI_SUCCESS) {
		return(lam_errfunc(comm, BLKMPIWAITANY, errstat));
	}

	lam_resetfunc_m(BLKMPIWAITANY);
	return(MPI_SUCCESS);
}
